from .base import ExordiumTests

from exordium.models import Artist, Album, Song, App, AlbumArt

class BasicAddTests(ExordiumTests):
    """
    Basic testing of the add() procedure under various circumstances.
    """

    ###
    ### Some methods called by the actual tests to get rid of some duplication
    ###

    def mp3_mode_test(self, mode):
        """
        Tests a simple addition of an mp3 of the given mode
        (abr/cbr/vbr) to the database, to ensure that that detection
        process is working properly.
        """
        self.add_mp3(artist='Artist', title='Title', basefile='silence-%s.mp3' % (mode.lower()))
        self.run_add()
        song = Song.objects.get(title='Title')
        self.assertEqual(song.mode, mode.upper())

    def mp3_year_test(self, year, yeartag):
        """
        Tests a simple addition of an mp3 to the database, using
        the specified year tag.
        """
        self.add_mp3(artist='Artist', title='Title', year=year, yeartag=yeartag)
        self.run_add()
        song = Song.objects.get()
        self.assertEqual(song.year, year)

    ###
    ### Actual tests follow
    ###

    def test_add_single_vbr_mp3(self):
        """
        Tests adding a single VBR mp3 to our library
        """
        self.mp3_mode_test('vbr')

    def test_add_single_cbr_mp3(self):
        """
        Tests adding a single CBR mp3 to our library
        """
        self.mp3_mode_test('cbr')

    def test_add_single_abr_mp3(self):
        """
        Tests adding a single ABR mp3 to our library
        """
        self.mp3_mode_test('abr')

    def test_add_with_no_various_artist(self):
        """
        Test what happens when we run an add without having a Various artist.
        It should be created.  Note that this won't actually happen unless
        we're also adding at least one real track.
        """
        self.assertEqual(Artist.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 0)
        self.assertEqual(Song.objects.count(), 0)
        ar = Artist.objects.get()
        self.assertEqual(ar.various, True)
        ar.delete()
        self.assertEqual(Artist.objects.count(), 0)

        self.add_mp3(artist='Artist', title='Title',
            filename='song.mp3')
        self.run_add()

        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Song.objects.count(), 1)
        ar = Artist.objects.get(name='Various')
        self.assertEqual(ar.various, True)

    def test_add_mp3_simple_tag_check(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_mp3(artist='Artist', title='Title', album='Album',
            year=1970, tracknum=1)
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.normtitle, 'title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')

    def test_add_classical_simple_tag_check(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_mp3(artist='Artist', title='Title', album='Album',
            group='Group', conductor='Conductor', composer='Composer',
            year=1970, tracknum=1)
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 5)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Group')
        self.assertEqual(artist.name, 'Group')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Conductor')
        self.assertEqual(artist.name, 'Conductor')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Composer')
        self.assertEqual(artist.name, 'Composer')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')
        self.assertEqual(song.group.name, 'Group')
        self.assertEqual(song.conductor.name, 'Conductor')
        self.assertEqual(song.composer.name, 'Composer')

    def test_add_ogg_simple_tag_check(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_ogg(artist='Artist', title='Title', album='Album',
            year=1970, tracknum=1)
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.normtitle, 'title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')

    def test_add_classical_simple_tag_check_ogg(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_ogg(artist='Artist', title='Title', album='Album',
            group='Group', conductor='Conductor', composer='Composer',
            year=1970, tracknum=1)
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 5)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Group')
        self.assertEqual(artist.name, 'Group')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Conductor')
        self.assertEqual(artist.name, 'Conductor')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Composer')
        self.assertEqual(artist.name, 'Composer')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')
        self.assertEqual(song.group.name, 'Group')
        self.assertEqual(song.conductor.name, 'Conductor')
        self.assertEqual(song.composer.name, 'Composer')

    def test_add_opus_simple_tag_check(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_opus(artist='Artist', title='Title', album='Album',
            year=1970, tracknum=1)
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.normtitle, 'title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')

    def test_add_classical_simple_tag_check_opus(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_opus(artist='Artist', title='Title', album='Album',
            group='Group', conductor='Conductor', composer='Composer',
            year=1970, tracknum=1)
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 5)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Group')
        self.assertEqual(artist.name, 'Group')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Conductor')
        self.assertEqual(artist.name, 'Conductor')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Composer')
        self.assertEqual(artist.name, 'Composer')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')
        self.assertEqual(song.group.name, 'Group')
        self.assertEqual(song.conductor.name, 'Conductor')
        self.assertEqual(song.composer.name, 'Composer')

    def test_add_m4a_simple_tag_check(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_m4a(artist='Artist', title='Title', album='Album',
            year=1970, tracknum=1)
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.normtitle, 'title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')

    def test_add_classical_simple_tag_check_m4a(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly
        """
        self.add_m4a(artist='Artist', title='Title', album='Album',
            composer='Composer', year=1970, tracknum=1)
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 3)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        artist = Artist.objects.get(name='Composer')
        self.assertEqual(artist.name, 'Composer')
        self.assertEqual(artist.prefix, '')

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.year, 1970)
        self.assertEqual(album.artist.name, 'Artist')

        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')
        self.assertEqual(song.composer.name, 'Composer')

    def test_add_mp3_total_track_tag_check(self):
        """
        Adds a single track to check for the alternate tracknum format
        where the maximum track count is included in the tag.
        """
        self.add_mp3(artist='Artist', title='Title', tracknum=1,
            maxtracks=5)
        self.run_add()

        song = Song.objects.get()
        self.assertEqual(song.tracknum, 1)

    def test_add_mp3_id3v23(self):
        """
        Adds a single track using ID3v2.3 encoding, rather than ID3v2.4.  The
        main thing we're checking here is for the year (whose tag changed between
        those two versions), but we'll be checking basically everything anyway.
        """
        self.add_mp3(artist='Artist', title='Title', album='Album',
            year=1970, tracknum=1, save_as_v23=True)
        self.run_add()
        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.year, 1970)
        self.assertEqual(song.tracknum, 1)
        self.assertEqual(song.album.name, 'Album')
        self.assertEqual(song.artist.name, 'Artist')

    def test_add_mp3_alternate_year_tdrl(self):
        """
        Adds a single track using the specific year tag TDRL, and verify
        that the year is populated properly.
        """
        self.mp3_year_test(1970, 'TDRL')

    def test_add_mp3_alternate_year_tdrc(self):
        """
        Adds a single track using the specific year tag TDRC, and verify
        that the year is populated properly.
        """
        self.mp3_year_test(1970, 'TDRC')

    def test_add_mp3_empty_track_tag(self):
        """
        Adds a single track with an empty tracknum field.  Note that
        mutagen refuses to write an actually-empty string to a tag,
        so we're using a space instead.  The important part is just
        that it's a value that fails when passed to int()
        """
        self.add_mp3(artist='Artist', title='Title', tracknum=' ')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        song = Song.objects.get()
        self.assertEqual(song.artist.name, 'Artist')
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.tracknum, 0)

    def test_add_mp3_empty_year_tag(self):
        """
        Adds a single track with an empty year field.  Mutagen refuses to
        write out invalid tags, so we've just constructed one to pass in.
        This technically also tests an empty tracknum field, making the
        previous test unnecessary, but whatever, we'll do both.
        """
        self.add_mp3(basefile='invalid-tags.mp3', apply_tags=False)
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        song = Song.objects.get()
        self.assertEqual(song.artist.name, 'Artist')
        self.assertEqual(song.title, 'Title')
        self.assertEqual(song.year, 0)
        self.assertEqual(song.tracknum, 0)

    def test_add_mp3_no_tags(self):
        """
        Test adding an mp3 file which has no tags specified
        """
        self.add_mp3(apply_tags=False)
        self.run_add_errors(error='Artist name not found')
        self.assertEqual(Song.objects.count(), 0)
        self.assertEqual(Artist.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 0)

    def test_add_mp3_no_title_tag(self):
        """
        Test adding an mp3 file which has no title tag specified
        """
        self.add_mp3(artist='Artist')
        self.run_add_errors(error='Title not found')
        self.assertEqual(Song.objects.count(), 0)
        self.assertEqual(Artist.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 0)

    def test_add_mp3_no_artist_tag(self):
        """
        Test adding an mp3 file which has no artist tag specified
        """
        self.add_mp3(title='Title', composer='Composer', album='Album')
        self.run_add_errors(error='Artist name not found')
        self.assertEqual(Song.objects.count(), 0)
        self.assertEqual(Artist.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 0)

    def test_add_without_filesystem_permissions(self):
        """
        Attempts adding a file that we don't actually have permission
        to read.  This does NOT produce any errors, because I intentionally
        mark some files as not readable by the user Django's running as,
        in order to hide them from Exordium.  It does get logged as a
        DEBUG message, but I won't bother looking for that.
        """
        self.add_mp3(artist='Artist', title='Title', album='Album',
            year=1970, tracknum=1, filename='song.mp3')
        self.set_file_unreadable('song.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 0)
        self.assertEqual(Artist.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 0)

    def test_add_invalid_filetype(self):
        """
        Attempts adding a file of a type we don't support (masquerading as
        one we do, with an invalid extension.  Using a .flac here)
        """
        self.add_file('silence.flac', 'song.ogg')
        self.run_add_errors(error='not yet understood')
        self.assertEqual(Song.objects.count(), 0)
        self.assertEqual(Artist.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 0)

    def test_add_various_reserved_name(self):
        """
        Attempts adding a file which uses the artist name "Various", which
        is currently a reserved value in Exordium.
        """
        self.add_mp3(artist='Various', title='Title',
            album='Album', filename='song.mp3')
        self.run_add_errors(error='Artist name "Various" is reserved')

        self.assertEqual(Song.objects.count(), 0)
        self.assertEqual(Artist.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 0)

    def test_add_mp3s_different_artist_case(self):
        """
        Adds two tracks by the same artist, but with different capitalization
        on the artist name.  Which version of the artist name gets stored is
        basically just dependent on whatever the app sees first.  I'm tempted
        to have it compare when it sees alternate cases and use the one with
        the most uppercase, but I think I'll just leave that to manual editing.
        """
        self.add_mp3(artist='Artist Name', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='artist name', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        # Note the mixed-case in the query, just checking that too.
        artist = Artist.objects.get(name__iexact='artist Name')
        self.assertEqual(artist.name.lower(), 'artist name')

    def test_add_song_different_composer_case(self):
        """
        Add a track with the same artist and composer but with different capitalization.
        Should only have one artist for it.
        """
        self.add_mp3(artist='Artist Name', composer='artist name', title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        artist = Artist.objects.get(name__iexact='artist Name')
        self.assertEqual(artist.name.lower(), 'artist name')
        song = Song.objects.get()
        self.assertEqual(song.artist.name.lower(), 'artist name')
        self.assertEqual(song.composer.name.lower(), 'artist name')

    def test_add_song_different_group_case(self):
        """
        Add a track with the same artist and group but with different capitalization.
        Should only have one artist for it.
        """
        self.add_mp3(artist='Artist Name', group='artist name', title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        artist = Artist.objects.get(name__iexact='artist Name')
        self.assertEqual(artist.name.lower(), 'artist name')
        song = Song.objects.get()
        self.assertEqual(song.artist.name.lower(), 'artist name')
        self.assertEqual(song.group.name.lower(), 'artist name')

    def test_add_song_different_conductor_case(self):
        """
        Add a track with the same artist and conductor but with different capitalization.
        Should only have one artist for it.
        """
        self.add_mp3(artist='Artist Name', conductor='artist name', title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        artist = Artist.objects.get(name__iexact='artist Name')
        self.assertEqual(artist.name.lower(), 'artist name')
        song = Song.objects.get()
        self.assertEqual(song.artist.name.lower(), 'artist name')
        self.assertEqual(song.conductor.name.lower(), 'artist name')

    def test_add_songs_different_group_cases(self):
        """
        Add two tracks with alternating artist/group names, both with different cases.
        """
        self.add_mp3(artist='Artist One', group='artist two', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist Two', group='artist one', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        artist = Artist.objects.get(name__iexact='artist One')
        self.assertEqual(artist.name.lower(), 'artist one')
        artist = Artist.objects.get(name__iexact='artist Two')
        self.assertEqual(artist.name.lower(), 'artist two')
        song = Song.objects.get(filename='song1.mp3')
        self.assertEqual(song.artist.name.lower(), 'artist one')
        self.assertEqual(song.group.name.lower(), 'artist two')
        song = Song.objects.get(filename='song2.mp3')
        self.assertEqual(song.artist.name.lower(), 'artist two')
        self.assertEqual(song.group.name.lower(), 'artist one')

    def test_add_songs_different_conductor_cases(self):
        """
        Add two tracks with alternating artist/conductor names, both with different cases.
        """
        self.add_mp3(artist='Artist One', conductor='artist two', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist Two', conductor='artist one', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        artist = Artist.objects.get(name__iexact='artist One')
        self.assertEqual(artist.name.lower(), 'artist one')
        artist = Artist.objects.get(name__iexact='artist Two')
        self.assertEqual(artist.name.lower(), 'artist two')
        song = Song.objects.get(filename='song1.mp3')
        self.assertEqual(song.artist.name.lower(), 'artist one')
        self.assertEqual(song.conductor.name.lower(), 'artist two')
        song = Song.objects.get(filename='song2.mp3')
        self.assertEqual(song.artist.name.lower(), 'artist two')
        self.assertEqual(song.conductor.name.lower(), 'artist one')

    def test_add_songs_different_composer_cases(self):
        """
        Add two tracks with alternating artist/composer names, both with different cases.
        """
        self.add_mp3(artist='Artist One', composer='artist two', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist Two', composer='artist one', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        artist = Artist.objects.get(name__iexact='artist One')
        self.assertEqual(artist.name.lower(), 'artist one')
        artist = Artist.objects.get(name__iexact='artist Two')
        self.assertEqual(artist.name.lower(), 'artist two')
        song = Song.objects.get(filename='song1.mp3')
        self.assertEqual(song.artist.name.lower(), 'artist one')
        self.assertEqual(song.composer.name.lower(), 'artist two')
        song = Song.objects.get(filename='song2.mp3')
        self.assertEqual(song.artist.name.lower(), 'artist two')
        self.assertEqual(song.composer.name.lower(), 'artist one')

    def test_add_mp3s_different_album_case(self):
        """
        Adds two tracks by the same artist and album, but with different
        capitalization on the album name.  Which version of the album name
        gets stored is basically just dependent on whatever the app sees first.
        """
        self.add_mp3(artist='Artist', album='Album Name',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist', album='album name',
            title='Title 2', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

        # Note the mixed-case in the query, just checking that too.
        album = Album.objects.get(name__iexact='album Name')
        self.assertEqual(album.name.lower(), 'album name')
        self.assertEqual(album.song_set.count(), 2)

    def test_add_mp3s_different_artist_and_album_case(self):
        """
        Adds two tracks by the same artist and album, but with different
        capitalization on the album and artist names.  Which version of the names
        gets stored is basically just dependent on whatever the app sees first.
        """
        self.add_mp3(artist='artist name', album='Album Name',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist Name', album='album name',
            title='Title 2', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

        # Note the mixed-case in the query, just checking that too.
        album = Album.objects.get(name__iexact='album Name')
        self.assertEqual(album.name.lower(), 'album name')
        self.assertEqual(album.song_set.count(), 2)
        artist = Artist.objects.get(name__iexact='artist Name')
        self.assertEqual(artist.name.lower(), 'artist name')

    def test_add_mp3s_differing_umlaut_artist(self):
        """
        Add two mp3s with the same artist but differing umlauts for the
        artist name.
        """
        self.add_mp3(artist='Umläut', album='Album',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Umlaut', album='Album',
            title='Title 2', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_song_differing_umlaut_group(self):
        """
        Add two mp3s with the same artist but differing umlauts for the
        artist name.
        """
        self.add_mp3(artist='Umläut', group='Umlaut', album='Album',
            title='Title 1', filename='song1.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)
        artist = Artist.objects.get(normname='umlaut')
        self.assertEqual(artist.name, 'Umläut')
        song = Song.objects.get()
        self.assertEqual(song.artist.normname, 'umlaut')
        self.assertEqual(song.group.normname, 'umlaut')

    def test_add_song_differing_umlaut_conductor(self):
        """
        Add two mp3s with the same artist but differing umlauts for the
        artist name.
        """
        self.add_mp3(artist='Umläut', conductor='Umlaut', album='Album',
            title='Title 1', filename='song1.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)
        artist = Artist.objects.get(normname='umlaut')
        self.assertEqual(artist.name, 'Umläut')
        song = Song.objects.get()
        self.assertEqual(song.artist.normname, 'umlaut')
        self.assertEqual(song.conductor.normname, 'umlaut')

    def test_add_song_differing_umlaut_composer(self):
        """
        Add two mp3s with the same artist but differing umlauts for the
        artist name.
        """
        self.add_mp3(artist='Umläut', composer='Umlaut', album='Album',
            title='Title 1', filename='song1.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)
        artist = Artist.objects.get(normname='umlaut')
        self.assertEqual(artist.name, 'Umläut')
        song = Song.objects.get()
        self.assertEqual(song.artist.normname, 'umlaut')
        self.assertEqual(song.composer.normname, 'umlaut')

    def test_add_songs_different_group_umlaut(self):
        """
        Add two tracks with alternating artist/group names, both with different umlauts.
        """
        self.add_mp3(artist='Umläut 1', group='Umlaut 2', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Umläut 2', group='Umlaut 1', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        song = Song.objects.get(filename='song1.mp3')
        self.assertEqual(song.artist.normname, 'umlaut 1')
        self.assertEqual(song.group.normname, 'umlaut 2')
        song = Song.objects.get(filename='song2.mp3')
        self.assertEqual(song.artist.normname, 'umlaut 2')
        self.assertEqual(song.group.normname, 'umlaut 1')

    def test_add_songs_different_conductor_umlaut(self):
        """
        Add two tracks with alternating artist/conductor names, both with different umlauts.
        """
        self.add_mp3(artist='Umläut 1', conductor='Umlaut 2', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Umläut 2', conductor='Umlaut 1', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        song = Song.objects.get(filename='song1.mp3')
        self.assertEqual(song.artist.normname, 'umlaut 1')
        self.assertEqual(song.conductor.normname, 'umlaut 2')
        song = Song.objects.get(filename='song2.mp3')
        self.assertEqual(song.artist.normname, 'umlaut 2')
        self.assertEqual(song.conductor.normname, 'umlaut 1')

    def test_add_songs_different_composer_umlaut(self):
        """
        Add two tracks with alternating artist/composer names, both with different umlauts.
        """
        self.add_mp3(artist='Umläut 1', composer='Umlaut 2', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Umläut 2', composer='Umlaut 1', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        song = Song.objects.get(filename='song1.mp3')
        self.assertEqual(song.artist.normname, 'umlaut 1')
        self.assertEqual(song.composer.normname, 'umlaut 2')
        song = Song.objects.get(filename='song2.mp3')
        self.assertEqual(song.artist.normname, 'umlaut 2')
        self.assertEqual(song.composer.normname, 'umlaut 1')

    def test_add_mp3s_differing_umlaut_album(self):
        """
        Add two mp3s with the same artist but differing umlauts for the
        album name.
        """
        self.add_mp3(artist='Umlaut', album='Albüm',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Umlaut', album='Album',
            title='Title 2', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_mp3s_differing_umlaut_album_and_artist(self):
        """
        Add two mp3s with the same artist but differing umlauts for the
        album and artist names.
        """
        self.add_mp3(artist='Umläut', album='Albüm',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Umlaut', album='Album',
            title='Title 2', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_mp3s_mismatched_japanese_artists(self):
        """
        Adds two files with different artist names using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        separate instead of collapsing them into a single artist.

        Characters taken from a search for "test" at google.co.jp,
        hopefully they are nothing offensive.  :)
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', album='Album',
            title='Title 1', filename='song1.mp3', path='Album1')
        self.add_mp3(artist='\u30AB\u30CA\u30C0', album='Album',
            title='Title 2', filename='song2.mp3', path='Album2')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        self.assertEqual(Album.objects.count(), 2)

    def test_add_mp3s_mismatched_japanese_artists_group(self):
        """
        Adds two files with different artist names using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        separate instead of collapsing them into a single artist.  Also tests
        the same functionality with groups.

        Characters taken from a search for "test" at google.co.jp,
        hopefully they are nothing offensive.  :)
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', group='\u30AB\u30CA\u30C0',
            album='Album', title='Title 1', filename='song1.mp3', path='Album1')
        self.add_mp3(artist='\u30AB\u30CA\u30C0', group='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 2', filename='song2.mp3', path='Album2')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        self.assertEqual(Album.objects.count(), 2)

    def test_add_mp3s_mismatched_japanese_artists_conductor(self):
        """
        Adds two files with different artist names using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        separate instead of collapsing them into a single artist.  Also tests
        the same functionality with conductors.

        Characters taken from a search for "test" at google.co.jp,
        hopefully they are nothing offensive.  :)
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', conductor='\u30AB\u30CA\u30C0',
            album='Album', title='Title 1', filename='song1.mp3', path='Album1')
        self.add_mp3(artist='\u30AB\u30CA\u30C0', conductor='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 2', filename='song2.mp3', path='Album2')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        self.assertEqual(Album.objects.count(), 2)

    def test_add_mp3s_mismatched_japanese_artists_composer(self):
        """
        Adds two files with different artist names using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        separate instead of collapsing them into a single artist.  Also tests
        the same functionality with composers.

        Characters taken from a search for "test" at google.co.jp,
        hopefully they are nothing offensive.  :)
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', composer='\u30AB\u30CA\u30C0',
            album='Album', title='Title 1', filename='song1.mp3', path='Album1')
        self.add_mp3(artist='\u30AB\u30CA\u30C0', composer='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 2', filename='song2.mp3', path='Album2')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 3)
        self.assertEqual(Album.objects.count(), 2)

    def test_add_mp3s_same_japanese_artists(self):
        """
        Adds two files with the same artist name using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        together.
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', album='Album',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', album='Album',
            title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_song_same_japanese_artist_group(self):
        """
        Adds one files with the same artist/group name using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        together.
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', group='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_song_same_japanese_artist_conductor(self):
        """
        Adds one files with the same artist/conductor name using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        together.
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', conductor='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_song_same_japanese_artist_composer(self):
        """
        Adds one files with the same artist/composer name using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        together.
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', composer='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_songs_same_japanese_artists_and_groups(self):
        """
        Adds two files with the same artist name using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        together.  Also test groups.
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', group='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', group='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_songs_same_japanese_artists_and_conductors(self):
        """
        Adds two files with the same artist name using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        together.  Also test conductors.
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', conductor='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', conductor='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_songs_same_japanese_artists_and_composers(self):
        """
        Adds two files with the same artist name using Japanese characters,
        to ensure that our artist-comparison normalization stuff keeps them
        together.  Also test composers.
        """
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', composer='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='\u81EA\u52D5\u8ABF', composer='\u81EA\u52D5\u8ABF',
            album='Album', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_mp3s_mismatched_japanese_albums(self):
        """
        Adds two files with different album names using Japanese characters,
        to ensure that our album-comparison normalization stuff keeps them
        separate instead of collapsing them into a single album.

        Characters taken from a search for "test" at google.co.jp,
        hopefully they are nothing offensive.  :)
        """
        self.add_mp3(artist='Artist', album='\u81EA\u52D5\u8ABF',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist', album='\u30AB\u30CA\u30C0',
            title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 2)

    def test_add_mp3s_mismatched_aesc_artist(self):
        """
        Adds two files with artist names which differ in that one uses an "æ" char
        and the other uses "ae".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='Mediæval', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Mediaeval', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_song_mismatched_aesc_artist_group(self):
        """
        Adds one files with artist/group names which differ in that one uses an
        "æ" char and the other uses "ae".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='Mediæval', group='Mediaeval',title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_song_mismatched_aesc_artist_conductor(self):
        """
        Adds one files with artist/conductor names which differ in that one uses an
        "æ" char and the other uses "ae".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='Mediæval', conductor='Mediaeval',title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_song_mismatched_aesc_artist_composer(self):
        """
        Adds one files with artist/composer names which differ in that one uses an
        "æ" char and the other uses "ae".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='Mediæval', composer='Mediaeval',title='Title 1', filename='song1.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_songs_mismatched_aesc_artist_and_group(self):
        """
        Adds two files with artist/group names which differ in that one uses an "æ" char
        and the other uses "ae".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='Mediæval', group='Mediaeval',title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Mediaeval', group='Mediæval',title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_songs_mismatched_aesc_artist_and_conductor(self):
        """
        Adds two files with artist/conductor names which differ in that one uses an "æ" char
        and the other uses "ae".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='Mediæval', conductor='Mediaeval',title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Mediaeval', conductor='Mediæval',title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_songs_mismatched_aesc_artist_and_composer(self):
        """
        Adds two files with artist/composer names which differ in that one uses an "æ" char
        and the other uses "ae".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='Mediæval', composer='Mediaeval',title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Mediaeval', composer='Mediæval',title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_mp3s_mismatched_aesc_album(self):
        """
        Adds two files with album names which differ in that one uses an "æ" char
        and the other uses "ae".  Should both normalize to the same album.
        """
        self.add_mp3(artist='Artist', album='Mediæval',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist', album='Mediaeval',
            title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_mp3s_mismatched_slashed_o_artist(self):
        """
        Adds two files with artist names which differ in that one uses an "ø" char
        and the other uses "o".  Should both normalize to the same artist.
        """
        self.add_mp3(artist='søster', title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='soster', title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)

    def test_add_mp3s_mismatched_slashed_o_album(self):
        """
        Adds two files with album names which differ in that one uses an "ø" char
        and the other uses "o".  Should both normalize to the same album.
        """
        self.add_mp3(artist='Artist', album='søster',
            title='Title 1', filename='song1.mp3')
        self.add_mp3(artist='Artist', album='soster',
            title='Title 2', filename='song2.mp3')
        self.run_add()
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Artist.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)

    def test_add_mp3_artist_prefix(self):
        """
        Adds a single track with an artist name "The Artist" to check for
        proper prefix handling.
        """
        self.add_mp3(artist='The Artist', title='Title')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

    def test_add_classical_simple_tag_check_prefixes(self):
        """
        Adds a single fully-tagged track and check that the resulting database
        objects are all populated properly.  This time with a prefix on all
        the artist fields.
        """
        self.add_mp3(artist='The Artist', title='Title', album='Album',
            group='The Group', conductor='The Conductor', composer='The Composer',
            year=1970, tracknum=1)
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 5)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

        artist = Artist.objects.get(name='Group')
        self.assertEqual(artist.name, 'Group')
        self.assertEqual(artist.prefix, 'The')

        artist = Artist.objects.get(name='Conductor')
        self.assertEqual(artist.name, 'Conductor')
        self.assertEqual(artist.prefix, 'The')

        artist = Artist.objects.get(name='Composer')
        self.assertEqual(artist.name, 'Composer')
        self.assertEqual(artist.prefix, 'The')

    def test_add_mp3_artist_prefix_double_spaces(self):
        """
        Adds a single track with an artist name "The  Artist" to check for
        proper prefix handling, with more space than necessary between the
        prefix and the actual artist name.
        """
        self.add_mp3(artist='The  Artist', title='Title')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

    def test_add_classical_prfix_double_spaces_full(self):
        """
        Adds a single fully-tagged track with some improper spaces between
        prefix and artist name - clean out the space and do the right thing.
        """
        self.add_mp3(artist='The  Artist', title='Title', group='The  Group',
                conductor='The  Conductor', composer='The  Composer')
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 5)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

        artist = Artist.objects.get(name='Group')
        self.assertEqual(artist.name, 'Group')
        self.assertEqual(artist.prefix, 'The')

        artist = Artist.objects.get(name='Conductor')
        self.assertEqual(artist.name, 'Conductor')
        self.assertEqual(artist.prefix, 'The')

        artist = Artist.objects.get(name='Composer')
        self.assertEqual(artist.name, 'Composer')
        self.assertEqual(artist.prefix, 'The')

    def test_add_mp3_with_exterior_spaces(self):
        """
        Adds a single track with both artist and album tags having extra
        spaces on the outside of the name.  Spaces should be stripped
        off.
        """
        self.add_mp3(artist=' Artist ', album=' Album ', title='Title')
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        album = Album.objects.get(name='Album')
        self.assertEqual(album.name, 'Album')

    def test_add_mp3_with_trailing_nulls(self):
        """
        Adds a single track with both artist and album tags having an
        extra NULL character at the end of the string.  The null char
        should be stripped out.
        """
        self.add_mp3(artist="Artist\x00", album="Album\x00", title="Title\x00")
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        album = Album.objects.get(name='Album')
        self.assertEqual(album.name, 'Album')
        song = Song.objects.get()
        self.assertEqual(song.title, 'Title')

    def test_add_mp3_normalize_title(self):
        """
        Test to ensure that song title normalization is working properly.
        """
        self.add_mp3(artist='Artist', title='Title: øÆ & ü')
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        song = Song.objects.get()
        self.assertEqual(song.title, 'Title: øÆ & ü')
        self.assertEqual(song.normtitle, 'title: oae and u')

    def test_add_mp3_no_album(self):
        """
        Adds an mp3 without an album to check that it's properly sorted
        into a 'Non-Album Tracks' album.
        """
        self.add_mp3(artist='Artist', title='Title')
        self.run_add()

        album_title = Album.miscellaneous_format_str % ('Artist')
        album = Album.objects.get()
        self.assertEqual(str(album), album_title)

    def test_add_mp3_no_album_two_stage(self):
        """
        Adds an mp3 without an album to check that it's properly sorted
        into a 'Non-Album Tracks' album, and then add a second using
        an umlaut to test out our matching abilities there
        """
        self.add_mp3(artist='Artist', title='Title', filename='song1.mp3')
        self.run_add()

        # Preliminary checks
        album_title = Album.miscellaneous_format_str % ('Artist')
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        album = Album.objects.get()
        self.assertEqual(album.miscellaneous, True)
        self.assertEqual(str(album), album_title)

        # Another add
        self.add_mp3(artist='Ärtist', title='Title 2', filename='song2.mp3')
        self.run_add()

        # Now the real checks
        album_title = Album.miscellaneous_format_str % ('Artist')
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        album = Album.objects.get()
        self.assertEqual(album.miscellaneous, True)
        self.assertEqual(str(album), album_title)
        self.assertEqual(album.song_set.count(), 2)

    def test_add_mp3_no_album_two_stage_aesc(self):
        """
        Adds an mp3 without an album to check that it's properly sorted
        into a 'Non-Album Tracks' album, and then add a second using
        an aesc to test out our matching abilities there.
        """
        self.add_mp3(artist='Aertist', title='Title', filename='song1.mp3')
        self.run_add()

        # Preliminary checks
        album_title = Album.miscellaneous_format_str % ('Aertist')
        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        album = Album.objects.get()
        self.assertEqual(album.miscellaneous, True)
        self.assertEqual(str(album), album_title)

        # Another add
        self.add_mp3(artist='Ærtist', title='Title 2', filename='song2.mp3')
        self.run_add()

        # Now the real checks
        album_title = Album.miscellaneous_format_str % ('Aertist')
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        album = Album.objects.get()
        self.assertEqual(album.miscellaneous, True)
        self.assertEqual(str(album), album_title)
        self.assertEqual(album.song_set.count(), 2)

    def test_add_mp3_two_song_album(self):
        """
        Adds two mp3s which should be in the same album.
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album', filename='1-title_1.mp3')
        self.add_mp3(artist='Artist', title='Title 2',
            album='Album', filename='2-title_2.mp3')
        self.run_add()

        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.artist.name, 'Artist')
        self.assertEqual(album.song_set.count(), 2)

    def test_add_mp3_two_alternating_prefix(self):
        """
        Add one mp3 with the artist name "The Artist"
        and then a second with the artist name "Artist".
        Both should be associated with the same base
        Artist name.
        """
        self.add_mp3(artist='The Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()
        self.add_mp3(artist='Artist', title='Title 2',
            album='Album 2', filename='album_2.mp3')
        self.run_add()

        for artist in Artist.objects.all():
            if artist.name != 'Various':
                self.assertEqual(artist.name, 'Artist')
                self.assertEqual(artist.prefix, 'The')
        for song in Song.objects.all():
            self.assertEqual(song.artist.name, 'Artist')

    def test_add_mp3_two_alternating_prefix_group(self):
        """
        Add one mp3 with the artist name "The Artist"
        and then a second with a group name "Artist".
        Both should be associated with the same base
        Artist name.
        """
        self.add_mp3(artist='The Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()
        self.add_mp3(artist='The Artist', group='Artist',
            title='Title 2', album='Album 2', filename='album_2.mp3')
        self.run_add()

        for artist in Artist.objects.all():
            if artist.name != 'Various':
                self.assertEqual(artist.name, 'Artist')
                self.assertEqual(artist.prefix, 'The')
        song = Song.objects.get(filename='album_2.mp3')
        self.assertEqual(song.group.name, 'Artist')

    def test_add_mp3_two_alternating_prefix_conductor(self):
        """
        Add one mp3 with the artist name "The Artist"
        and then a second with a conductor name "Artist".
        Both should be associated with the same base
        Artist name.
        """
        self.add_mp3(artist='The Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()
        self.add_mp3(artist='The Artist', conductor='Artist',
            title='Title 2', album='Album 2', filename='album_2.mp3')
        self.run_add()

        for artist in Artist.objects.all():
            if artist.name != 'Various':
                self.assertEqual(artist.name, 'Artist')
                self.assertEqual(artist.prefix, 'The')
        song = Song.objects.get(filename='album_2.mp3')
        self.assertEqual(song.conductor.name, 'Artist')

    def test_add_mp3_two_alternating_prefix_composer(self):
        """
        Add one mp3 with the artist name "The Artist"
        and then a second with a composer name "Artist".
        Both should be associated with the same base
        Artist name.
        """
        self.add_mp3(artist='The Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()
        self.add_mp3(artist='The Artist', composer='Artist',
            title='Title 2', album='Album 2', filename='album_2.mp3')
        self.run_add()

        for artist in Artist.objects.all():
            if artist.name != 'Various':
                self.assertEqual(artist.name, 'Artist')
                self.assertEqual(artist.prefix, 'The')
        song = Song.objects.get(filename='album_2.mp3')
        self.assertEqual(song.composer.name, 'Artist')

    def test_add_mp3_two_alternating_prefix_reverse(self):
        """
        Add one mp3 with the artist name "Artist" and
        then a second one with the artist name "The Artist".
        The artist record should be updated to have a
        prefix.
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        self.add_mp3(artist='The Artist', title='Title 2',
            album='Album 2', filename='album_2.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

    def test_add_mp3_two_alternating_prefix_reverse_by_group(self):
        """
        Add one mp3 with the artist name "Artist" and
        then a second one with a group name "The Artist".
        The artist record should be updated to have a
        prefix.
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        self.add_mp3(artist='Artist', title='Title 2',
            group='The Artist',
            album='Album 2', filename='album_2.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

    def test_add_mp3_two_alternating_prefix_reverse_by_conductor(self):
        """
        Add one mp3 with the artist name "Artist" and
        then a second one with a conductor name "The Artist".
        The artist record should be updated to have a
        prefix.
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        self.add_mp3(artist='Artist', title='Title 2',
            conductor='The Artist',
            album='Album 2', filename='album_2.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

    def test_add_mp3_two_alternating_prefix_reverse_by_composer(self):
        """
        Add one mp3 with the artist name "Artist" and
        then a second one with a composer name "The Artist".
        The artist record should be updated to have a
        prefix.
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album 1', filename='album_1.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, '')

        self.add_mp3(artist='Artist', title='Title 2',
            composer='The Artist',
            album='Album 2', filename='album_2.mp3')
        self.run_add()

        artist = Artist.objects.get(name='Artist')
        self.assertEqual(artist.name, 'Artist')
        self.assertEqual(artist.prefix, 'The')

    def test_add_various_artists_album(self):
        """
        Test adding a various-artists album.
        """
        tracks = [
            ('1-title_1.mp3', 'Artist 1', 'Title 1', 1),
            ('2-title_2.mp3', 'Artist 2', 'Title 2', 2),
        ]
        for (filename, artist, title, tracknum) in tracks:
            self.add_mp3(artist=artist, title=title, album='Album',
                tracknum=tracknum, filename=filename)
        self.run_add()

        # First check that we have three total artists, and that
        # they're what we expect
        self.assertEqual(Artist.objects.count(), 3)
        va = Artist.objects.get(name='Various')
        self.assertEqual(va.name, 'Various')
        a1 = Artist.objects.get(name='Artist 1')
        self.assertEqual(a1.name, 'Artist 1')
        a2 = Artist.objects.get(name='Artist 2')
        self.assertEqual(a2.name, 'Artist 2')

        # Now check that there's just one album and that IT is
        # how we expect.
        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get()
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.artist.name, 'Various')
        self.assertEqual(album.song_set.count(), 2)

        # Aaand check the individual songs
        for (filename, artist, title, tracknum) in tracks:
            song = Song.objects.get(filename=filename)
            self.assertEqual(song.title, title)
            self.assertEqual(song.artist.name, artist)
            self.assertEqual(song.album.name, 'Album')
            self.assertEqual(song.album.artist.name, 'Various')
            self.assertEqual(song.tracknum, tracknum)

    def test_add_va_album_and_normal_album(self):
        """
        Create a Various Artists album, and at the same time another
        album belonging to one of the artists on the VA comp.
        """
        tracks = [
            ('Various', '1-title_1.mp3', 'Artist 1', 'Title 1', 1),
            ('Various', '2-title_2.mp3', 'Artist 2', 'Title 2', 2),
            ('Album', '1-first.mp3', 'Artist 1', 'First', 1),
            ('Album', '2-second.mp3', 'Artist 1', 'Second', 2),
        ]
        for (path, filename, artist, title, tracknum) in tracks:
            self.add_mp3(path=path, filename=filename, artist=artist,
                title=title, tracknum=tracknum, album=path)
        self.run_add()

        # First check that we have three total artists, and that
        # they're what we expect
        self.assertEqual(Artist.objects.count(), 3)
        va = Artist.objects.get(name='Various')
        self.assertEqual(va.name, 'Various')
        self.assertEqual(va.album_set.count(), 1)
        a1 = Artist.objects.get(name='Artist 1')
        self.assertEqual(a1.name, 'Artist 1')
        self.assertEqual(a1.album_set.count(), 1)
        a2 = Artist.objects.get(name='Artist 2')
        self.assertEqual(a2.name, 'Artist 2')
        self.assertEqual(a2.album_set.count(), 0)

        # Now check for two albums
        self.assertEqual(Album.objects.count(), 2)
        various = Album.objects.get(name='Various')
        self.assertEqual(various.name, 'Various')
        self.assertEqual(various.artist.name, 'Various')
        self.assertEqual(various.song_set.count(), 2)
        single = Album.objects.get(name='Album')
        self.assertEqual(single.name, 'Album')
        self.assertEqual(single.artist.name, 'Artist 1')
        self.assertEqual(single.song_set.count(), 2)

        # Now check the individual album tracks
        for song in various.song_set.all():
            self.assertNotEqual(song.artist.name, 'Various')
            self.assertEqual(song.album.name, 'Various')
            self.assertEqual(song.album.artist.name, 'Various')
        for song in single.song_set.all():
            self.assertEqual(song.artist.name, 'Artist 1')
            self.assertEqual(song.album.name, 'Album')
            self.assertEqual(song.album.artist.name, 'Artist 1')

    def test_add_normal_track_to_normal_album(self):
        """
        Create a single-artist album and then add another track to the
        album (by the same artist).
        """
        tracks = [
            ('Album', '1-first.mp3', 'Artist 1', 'First', 1),
            ('Album', '2-second.mp3', 'Artist 1', 'Second', 2),
        ]
        for (path, filename, artist, title, tracknum) in tracks:
            self.add_mp3(path=path, filename=filename, artist=artist,
                title=title, tracknum=tracknum, album=path)
        self.run_add()

        # Just verify that the album is Artist 1
        album = Album.objects.get(name='Album')
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.artist.name, 'Artist 1')

        # Now add the new track
        self.add_mp3(path='Album', filename='3-third.mp3',
            artist='Artist 1', title='Third', tracknum=3,
            album='Album')
        self.run_add()

        # Verify that we only have one album, and that it's still Artist 1
        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get(name='Album')
        self.assertEqual(album.song_set.count(), 3)
        self.assertEqual(album.artist.name, 'Artist 1')

    def test_add_va_track_to_va_album(self):
        """
        Create a various-artist album and then add another track to the
        album.
        """
        tracks = [
            ('Album', '1-first.mp3', 'Artist 1', 'First', 1),
            ('Album', '2-second.mp3', 'Artist 2', 'Second', 2),
        ]
        for (path, filename, artist, title, tracknum) in tracks:
            self.add_mp3(path=path, filename=filename, artist=artist,
                title=title, tracknum=tracknum, album=path)
        self.run_add()

        # Just verify that the album is Various
        album = Album.objects.get(name='Album')
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.artist.name, 'Various')

        # Now add the new track
        self.add_mp3(path='Album', filename='3-third.mp3',
            artist='Artist 3', title='Third', tracknum=3,
            album='Album')
        self.run_add()

        # Verify that we only have one album, and that it's still VA
        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get(name='Album')
        self.assertEqual(album.song_set.count(), 3)
        self.assertEqual(album.artist.name, 'Various')

    def test_add_va_track_to_non_va_album(self):
        """
        Create a regular single-artist album and then add a track to the
        album which turns it into a Various Artists album.
        """
        tracks = [
            ('Album', '1-first.mp3', 'Artist 1', 'First', 1),
            ('Album', '2-second.mp3', 'Artist 1', 'Second', 2),
        ]
        for (path, filename, artist, title, tracknum) in tracks:
            self.add_mp3(path=path, filename=filename, artist=artist,
                title=title, tracknum=tracknum, album=path)
        self.run_add()

        # Just verify that the album is Artist 1 quick
        self.assertEqual(Song.objects.count(), 2)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        album = Album.objects.get(name='Album')
        self.assertEqual(album.name, 'Album')
        self.assertEqual(album.artist.name, 'Artist 1')

        # Now add the new track
        self.add_mp3(path='Album', filename='3-third.mp3',
            artist='Artist 2', title='Third', tracknum=3,
            album='Album')
        self.run_add()

        # Verify that we only have one album, and that it's changed
        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get(name='Album')
        self.assertEqual(album.song_set.count(), 3)
        self.assertEqual(album.artist.name, 'Various')

    def test_add_two_regular_albums_with_same_album_name(self):
        """
        Test behavior when two regular albums are added with the same
        album name.  This is a subpar situation and may result in
        problems down the line, but for now the expected behavior
        is that there will be a single regular album containing tracks
        from both directories.
        """
        tracks = [
            ('al1', 'Album', '1-first.mp3', 'Artist 1', 'First', 1),
            ('al1', 'Album', '2-second.mp3', 'Artist 1', 'Second', 2),
            ('al2', 'Album', '1-first.mp3', 'Artist 1', 'First', 1),
            ('al2', 'Album', '2-second.mp3', 'Artist 1', 'Second', 2),
        ]
        for (path, album, filename, artist, title, tracknum) in tracks:
            self.add_mp3(path=path, filename=filename, artist=artist,
                title=title, tracknum=tracknum, album=album)
        self.run_add()

        # Some simple checks
        self.assertEqual(Song.objects.count(), 4)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        album = Album.objects.get()
        self.assertEqual(album.song_set.count(), 4)
        for song in Song.objects.all():
            self.assertEqual(song.album.name, 'Album')
            self.assertEqual(song.album.artist.name, 'Artist 1')
            self.assertEqual(song.artist.name, 'Artist 1')

    def test_add_two_va_albums_with_same_album_name(self):
        """
        Test behavior when two V/A albums are added with the same
        album name.  This is a subpar situation and may result in
        problems down the line, but for now the expected behavior
        is that there will be a single V/A album containing tracks
        from both directories.
        """
        tracks = [
            ('va1', 'Album', '1-first.mp3', 'Artist 1', 'First', 1),
            ('va1', 'Album', '2-second.mp3', 'Artist 2', 'Second', 2),
            ('va2', 'Album', '1-first.mp3', 'Artist 3', 'First', 1),
            ('va2', 'Album', '2-second.mp3', 'Artist 4', 'Second', 2),
        ]
        for (path, album, filename, artist, title, tracknum) in tracks:
            self.add_mp3(path=path, filename=filename, artist=artist,
                title=title, tracknum=tracknum, album=album)
        self.run_add()

        # Some simple checks
        self.assertEqual(Song.objects.count(), 4)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 5)
        album = Album.objects.get()
        self.assertEqual(album.song_set.count(), 4)
        for song in Song.objects.all():
            self.assertEqual(song.album.name, 'Album')
            self.assertEqual(song.album.artist.name, 'Various')

    def test_add_live_album(self):
        """
        Simple test to add an album with a title which will mark it as
        being "live"
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='2016-09-20 - Live at Somewhere',
            filename='song1.mp3')
        self.run_add()

        self.assertEqual(Song.objects.count(), 1)
        self.assertEqual(Album.objects.count(), 1)
        self.assertEqual(Artist.objects.count(), 2)
        album = Album.objects.get()
        self.assertEqual(album.live, True)

class BasicAddAsUpdateTests(BasicAddTests):
    """
    This is a bit of nonsense.  Basically, all tests for add() should be
    repeated for update(), since add() is technically a subset of update().
    Rather than rewriting everything, we're just subclassing and
    overriding the ``run_add()`` method so that all calls to ``run_add()``
    end up doing an update instead.
    """

    def run_add(self):
        """
        Runs an ``update`` operation on our library while pretending to be
        ``add``, and checks for errors.
        """
        return self.assertNoErrors(list(App.update()))

    def run_add_errors(self, errors_min=1, error=None):
        """
        Runs an ``update`` operation on our library while pretending to be
        ``add``, and ensures that there's at least one error
        """
        return self.assertErrors(list(App.update()), errors_min, error=error)




from django.test import TestCase

from exordium.models import Artist, Album, Song, App, AlbumArt

class AlbumZipfileErrorModelTests(TestCase):
    """
    Tests for our App.AlbumZipfileError exception, since we don't have
    a way to legitimately generating one of those.  Mostly just checking
    that the stored exception works all right.
    """

    def test_orig_exception(self):
        """
        Tests that our orig_exception attribute works as-expected
        """
        with self.assertRaises(App.AlbumZipfileError) as cm:
            try:
                a = int('a')
            except ValueError as e:
                raise App.AlbumZipfileError(e)
        self.assertEqual(type(cm.exception.orig_exception), type(ValueError()))




from django.test import TestCase, override_settings
from django.utils import timezone

from django.contrib.auth.models import User

from dynamic_preferences.registries import global_preferences_registry

import os
import shutil
import pathlib
import datetime
import tempfile

from mutagen.id3 import ID3, TIT2, TALB, TPE1, TDRC, TRCK, TDRL, TPE2, TPE3, TCOM
from mutagen.oggvorbis import OggVorbis
from mutagen.oggopus import OggOpus
from mutagen.mp4 import MP4

from exordium.models import Artist, Album, Song, App, AlbumArt

# These two imports are just here in case we want to examine SQL while running tests.
# If so, set "settings.DEBUG = True" in the test and then use connection.queries
#from django.conf import settings
#from django.db import connection

class ExordiumTests(TestCase):
    """
    Custom TestCase class for Exordium.  Includes our ``initial_data``
    fixture and sets up a pretend library under ``testdata``.  Uses
    the sample mp3 files in ``testdata`` as the basis for all the files
    that we'll be testing with.
    """

    fixtures = ['initial_data.json']

    testdata_path = os.path.join(os.path.dirname(__file__), '..', 'testdata')
    library_path = None

    prefs = None

    def setUp(self):
        """
        Run automatically at the start of any test.  Ensure that there's no
        library hanging around in our testdata dir, ensure that our base
        testing files exist, and then set up the base library path.
        """
        for filename in ['silence-abr.mp3', 'silence-cbr.mp3', 'silence-vbr.mp3', 'invalid-tags.mp3',
                'silence.ogg', 'silence.m4a', 'silence.flac', 'silence.opus',
                'cover_400.jpg', 'cover_400.gif', 'cover_400.png', 'cover_400.tif', 'cover_100.jpg']:
            if not os.path.exists(os.path.join(self.testdata_path, filename)):  # pragma: no cover
                raise Exception('Required testing file "%s" does not exist!' % (filename))
        self.library_path = tempfile.mkdtemp()
        self.prefs = global_preferences_registry.manager()
        self.prefs['exordium__base_path'] = self.library_path
        self.prefs['exordium__media_url_html5'] = 'http://testserver-media/html5music'
        self.prefs['exordium__media_url_m3u'] = 'http://testserver-media/m3umusic'

        # We have one test which alters the following value, which
        # will stay changed between tests unless we restore it.
        self.saved_album_size = AlbumArt.resolutions[AlbumArt.SZ_ALBUM]

    def tearDown(self):
        """
        Run automatically at the test conclusion.  Get rid of our base
        library.
        """
        shutil.rmtree(self.library_path)
        AlbumArt.resolutions[AlbumArt.SZ_ALBUM] = self.saved_album_size

    def check_library_filename(self, filename):
        """
        Checks to make sure our passed-in filename isn't trying to break
        out of our library.  This is a bit silly since only these tests
        will ever be passing in paths, and we'd have to be attempting to
        do nasty things to ourselves, but this should help prevent any
        unintentional blunders, at least.

        Returns our full filename, with library path, since we basically
        always want that after doing this check.
        """
        if len(filename) < 3 or '..' in filename or filename[0] == '/':
            raise Exception('Given filename "%s" is invalid' % (filename))
        return os.path.join(self.library_path, filename)

    def add_file(self, basefile, filename, path=''):
        """
        Adds an arbitrary datafile somewhere in our library.  Returns the
        full filename.
        """
        if path != '' and ('..' in path or path[0] == '/'):
            raise Exception('Given path "%s" is invalid' % (path))

        if '/' in basefile or len(basefile) < 3 or '.' not in basefile:
            raise Exception('Invalid basefile name: %s' % (basefile))

        src_filename = os.path.join(self.testdata_path, basefile)
        if not os.path.exists(src_filename):
            raise Exception('Source filename %s is not found' % (src_filename))

        full_path = os.path.join(self.library_path, path)
        full_filename = os.path.join(full_path, filename)
        os.makedirs(full_path, exist_ok=True)
        shutil.copyfile(src_filename, full_filename)
        self.assertEqual(os.path.exists(full_filename), True)

        return full_filename

    def add_mp3(self, path='', filename='file.mp3', artist='', album='',
            title='', tracknum=0, maxtracks=None, year=0, yeartag='TDRC',
            group='', conductor='', composer='',
            basefile='silence-vbr.mp3', save_as_v23=False,
            apply_tags=True):
        """
        Adds a new mp3 with the given parameters to our library.

        Pass in ``save_as_v23`` as ``True`` to have the file save with an ID3v2.3
        tag, instead of ID3v2.4.  One of the main tag-level changes which
        will happen there is conversion of the year tag to TYER, which
        we'll otherwise not be specifying directly.  ``yeartag`` is effectively
        ignored when ``save_as_v23`` is True.

        Pass in ``False`` for ``apply_tags`` to only use whatever tags happen to
        be present in the source basefile.

        Returns the full filename of the added file.
        """

        full_filename = self.add_file(basefile, filename, path=path)

        # Finish here if we've been told to.
        if not apply_tags:
            return full_filename

        # Apply the tags as specified
        tags = ID3()
        tags.add(TPE1(encoding=3, text=artist))
        tags.add(TALB(encoding=3, text=album))
        tags.add(TIT2(encoding=3, text=title))

        if group != '':
            tags.add(TPE2(encoding=3, text=group))
        if conductor != '':
            tags.add(TPE3(encoding=3, text=conductor))
        if composer != '':
            tags.add(TCOM(encoding=3, text=composer))

        if maxtracks is None:
            tags.add(TRCK(encoding=3, text=str(tracknum)))
        else:
            tags.add(TRCK(encoding=3, text='%s/%s' % (tracknum, maxtracks)))

        if yeartag == 'TDRL':
            tags.add(TDRL(encoding=3, text=str(year)))
        elif yeartag == 'TDRC':
            tags.add(TDRC(encoding=3, text=str(year)))
        else:
            raise Exception('Unknown year tag specified: %s' % (yeartag))

        # Convert to ID3v2.3 if requested.
        if save_as_v23:
            tags.update_to_v23()

        # Save to our filename
        tags.save(full_filename)

        # Return
        return full_filename

    def bump_mtime(self, starting_mtime, filename=None, full_filename=None):
        """
        Given the specified `filename` or `full_filename`, ensure that its
        on-disk mtime is higher than `starting_mtime`.
        """

        # Figure out the full filename and make sure it exists
        if full_filename is None:
            full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)

        # Now grab the new mtime and make sure that it's at least 1sec
        # higher than `starting_mtime`
        stat_result = os.stat(full_filename)
        ending_mtime = int(stat_result.st_mtime)
        if starting_mtime == ending_mtime:
            new_mtime = ending_mtime + 1
            os.utime(full_filename, times=(stat_result.st_atime, new_mtime))

    def update_mp3(self, filename, artist=None, album=None,
            title=None, tracknum=None, maxtracks=None, year=None,
            group=None, conductor=None, composer=None):
        """
        Updates an on-disk mp3 with the given tag data.  Any passed-in
        variable set to None will be ignored.  It's possible there could
        be some problems with ID3v2.3 vs. ID3v2.4 tags in here - I don't
        know if mutagen does an auto-convert.  I think it might.

        If group/conductor/composer is a blank string, those fields will
        be completely removed from the file.  Any of the other fields set
        to blank will leave the tag in place.

        Will ensure that the file's mtime is updated.
        """

        full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)

        starting_mtime = int(os.stat(full_filename).st_mtime)

        tags = ID3(full_filename)

        if artist is not None:
            tags.delall('TPE1')
            tags.add(TPE1(encoding=3, text=artist))

        if album is not None:
            tags.delall('TALB')
            tags.add(TALB(encoding=3, text=album))

        if title is not None:
            tags.delall('TIT2')
            tags.add(TIT2(encoding=3, text=title))

        if group is not None:
            tags.delall('TPE2')
            if group != '':
                tags.add(TPE2(encoding=3, text=group))

        if conductor is not None:
            tags.delall('TPE3')
            if conductor != '':
                tags.add(TPE3(encoding=3, text=conductor))

        if composer is not None:
            tags.delall('TCOM')
            if composer != '':
                tags.add(TCOM(encoding=3, text=composer))

        if tracknum is not None:
            tags.delall('TRCK')
            if maxtracks is None:
                tags.add(TRCK(encoding=3, text=str(tracknum)))
            else:
                tags.add(TRCK(encoding=3, text='%s/%s' % (tracknum, maxtracks)))

        if year is not None:
            tags.delall('TDRC')
            tags.delall('TDRL')
            tags.delall('TYER')
            tags.add(TDRC(encoding=3, text=str(year)))

        # Save
        tags.save()

        # Check on mtime update and manually fix it if it's not updated
        self.bump_mtime(starting_mtime, full_filename=full_filename)

    def add_ogg(self, path='', filename='file.ogg', artist='', album='',
            title='', tracknum=None, year=None, group='', conductor='', composer='',
            basefile='silence.ogg', apply_tags=True):
        """
        Adds a new ogg with the given parameters to our library.

        Pass in ``False`` for ``apply_tags`` to only use whatever tags happen to
        be present in the source basefile.
        """

        full_filename = self.add_file(basefile, filename, path=path)

        # Finish here if we've been told to.
        if not apply_tags:
            return

        # Apply the tags as specified
        tags = OggVorbis(full_filename)
        tags['ARTIST'] = artist
        tags['ALBUM'] = album
        tags['TITLE'] = title

        if group != '':
            tags['ENSEMBLE'] = group
        if conductor != '':
            tags['CONDUCTOR'] = conductor
        if composer != '':
            tags['COMPOSER'] = composer
        if tracknum is not None:
            tags['TRACKNUMBER'] = str(tracknum)
        if year is not None:
            tags['DATE'] = str(year)

        # Save to our filename
        tags.save()

    def update_ogg(self, filename, artist=None, album=None,
            title=None, tracknum=None, year=None,
            group=None, conductor=None, composer=None):
        """
        Updates an on-disk ogg with the given tag data.  Any passed-in
        variable set to None will be ignored.

        If group/conductor/composer is a blank string, those fields will
        be completely removed from the file.  Any of the other fields set
        to blank will leave the tag in place.

        Will ensure that the file's mtime is updated.
        """

        full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)

        starting_mtime = int(os.stat(full_filename).st_mtime)

        tags = OggVorbis(full_filename)

        if artist is not None:
            try:
                del tags['ARTIST']
            except KeyError:
                pass
            tags['ARTIST'] = artist

        if album is not None:
            try:
                del tags['ALBUM']
            except KeyError:
                pass
            tags['ALBUM'] = album

        if title is not None:
            try:
                del tags['TITLE']
            except KeyError:
                pass
            tags['TITLE'] = title

        if group is not None:
            try:
                del tags['ENSEMBLE']
            except KeyError:
                pass
            if group != '':
                tags['ENSEMBLE'] = group

        if conductor is not None:
            try:
                del tags['CONDUCTOR']
            except KeyError:
                pass
            if conductor != '':
                tags['CONDUCTOR'] = conductor

        if composer is not None:
            try:
                del tags['COMPOSER']
            except KeyError:
                pass
            if composer != '':
                tags['COMPOSER'] = composer

        if tracknum is not None:
            try:
                del tags['TRACKNUMBER']
            except KeyError:
                pass
            tags['TRACKNUMBER'] = str(tracknum)

        if year is not None:
            try:
                del tags['DATE']
            except KeyError:
                pass
            try:
                del tags['YEAR']
            except KeyError:
                pass
            tags['DATE'] = str(year)

        # Save
        tags.save()

        # Check on mtime update and manually fix it if it's not updated
        stat_result = os.stat(full_filename)
        ending_mtime = int(stat_result.st_mtime)
        if starting_mtime == ending_mtime:
            new_mtime = ending_mtime + 1
            os.utime(full_filename, times=(stat_result.st_atime, new_mtime))

    def add_opus(self, path='', filename='file.opus', artist='', album='',
            title='', tracknum=None, year=None, group='', conductor='', composer='',
            basefile='silence.opus', apply_tags=True):
        """
        Adds a new ogg opus file with the given parameters to our library.

        Pass in ``False`` for ``apply_tags`` to only use whatever tags happen to
        be present in the source basefile.
        """

        full_filename = self.add_file(basefile, filename, path=path)

        # Finish here if we've been told to.
        if not apply_tags:
            return

        # Apply the tags as specified
        tags = OggOpus(full_filename)
        tags['ARTIST'] = artist
        tags['ALBUM'] = album
        tags['TITLE'] = title

        if group != '':
            tags['ENSEMBLE'] = group
        if conductor != '':
            tags['CONDUCTOR'] = conductor
        if composer != '':
            tags['COMPOSER'] = composer
        if tracknum is not None:
            tags['TRACKNUMBER'] = str(tracknum)
        if year is not None:
            tags['DATE'] = str(year)

        # Save to our filename
        tags.save()

    def update_opus(self, filename, artist=None, album=None,
            title=None, tracknum=None, year=None,
            group=None, conductor=None, composer=None):
        """
        Updates an on-disk ogg opus file with the given tag data.  Any passed-in
        variable set to None will be ignored.

        If group/conductor/composer is a blank string, those fields will
        be completely removed from the file.  Any of the other fields set
        to blank will leave the tag in place.

        Will ensure that the file's mtime is updated.
        """

        full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)

        starting_mtime = int(os.stat(full_filename).st_mtime)

        tags = OggOpus(full_filename)

        if artist is not None:
            try:
                del tags['ARTIST']
            except KeyError:
                pass
            tags['ARTIST'] = artist

        if album is not None:
            try:
                del tags['ALBUM']
            except KeyError:
                pass
            tags['ALBUM'] = album

        if title is not None:
            try:
                del tags['TITLE']
            except KeyError:
                pass
            tags['TITLE'] = title

        if group is not None:
            try:
                del tags['ENSEMBLE']
            except KeyError:
                pass
            if group != '':
                tags['ENSEMBLE'] = group

        if conductor is not None:
            try:
                del tags['CONDUCTOR']
            except KeyError:
                pass
            if conductor != '':
                tags['CONDUCTOR'] = conductor

        if composer is not None:
            try:
                del tags['COMPOSER']
            except KeyError:
                pass
            if composer != '':
                tags['COMPOSER'] = composer

        if tracknum is not None:
            try:
                del tags['TRACKNUMBER']
            except KeyError:
                pass
            tags['TRACKNUMBER'] = str(tracknum)

        if year is not None:
            try:
                del tags['DATE']
            except KeyError:
                pass
            try:
                del tags['YEAR']
            except KeyError:
                pass
            tags['DATE'] = str(year)

        # Save
        tags.save()

        # Check on mtime update and manually fix it if it's not updated
        stat_result = os.stat(full_filename)
        ending_mtime = int(stat_result.st_mtime)
        if starting_mtime == ending_mtime:
            new_mtime = ending_mtime + 1
            os.utime(full_filename, times=(stat_result.st_atime, new_mtime))

    def add_m4a(self, path='', filename='file.m4a', artist='', album='',
            title='', tracknum=None, year=None, composer='',
            basefile='silence.m4a', apply_tags=True):
        """
        Adds a new m4a with the given parameters to our library.

        Note that m4a tagging doesn't seem to support conductor or group/ensemble. 

        Pass in ``False`` for ``apply_tags`` to only use whatever tags happen to
        be present in the source basefile.
        """

        full_filename = self.add_file(basefile, filename, path=path)

        # Finish here if we've been told to.
        if not apply_tags:
            return

        # Apply the tags as specified
        tags = MP4(full_filename)
        tags['\xa9ART'] = artist
        tags['\xa9alb'] = album
        tags['\xa9nam'] = title

        if composer != '':
            tags['\xa9wrt'] = composer
        if tracknum is not None:
            tags['trkn'] = [(tracknum, 0)]
        if year is not None:
            tags['\xa9day'] = str(year)

        # Save to our filename
        tags.save()

    def update_m4a(self, filename, artist=None, album=None,
            title=None, tracknum=None, year=None,
            composer=None):
        """
        Updates an on-disk ogg with the given tag data.  Any passed-in
        variable set to None will be ignored.

        Note again that m4a tagging apparently doesn't support conductor or
        group/ensemble.

        If composer is a blank string, that field will be completely
        removed from the file.  Any of the other fields set to blank will
        leave the tag in place.

        Will ensure that the file's mtime is updated.
        """

        full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)

        starting_mtime = int(os.stat(full_filename).st_mtime)

        tags = MP4(full_filename)

        if artist is not None:
            try:
                del tags['\xa9ART']
            except KeyError:
                pass
            tags['\xa9ART'] = artist

        if album is not None:
            try:
                del tags['\xa9alb']
            except KeyError:
                pass
            tags['\xa9alb'] = album

        if title is not None:
            try:
                del tags['\xa9nam']
            except KeyError:
                pass
            tags['\xa9nam'] = title

        if composer is not None:
            try:
                del tags['\xa9wrt']
            except KeyError:
                pass
            if composer != '':
                tags['\xa9wrt'] = composer

        if tracknum is not None:
            try:
                del tags['trkn']
            except KeyError:
                pass
            tags['trkn'] = [(tracknum, 0)]

        if year is not None:
            try:
                del tags['\xa9day']
            except KeyError:
                pass
            tags['\xa9day'] = str(year)

        # Save
        tags.save()

        # Check on mtime update and manually fix it if it's not updated
        stat_result = os.stat(full_filename)
        ending_mtime = int(stat_result.st_mtime)
        if starting_mtime == ending_mtime:
            new_mtime = ending_mtime + 1
            os.utime(full_filename, times=(stat_result.st_atime, new_mtime))

    def delete_file(self, filename):
        """
        Deletes the given file from our fake library
        """
        full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)

        os.unlink(full_filename)

        self.assertEqual(os.path.exists(full_filename), False)

    def rename_file(self, filename, new_filename):
        """
        Renames the given file in our fake library to a new filename (in
        the same directory)
        """
        full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)
        full_new_filename = self.check_library_filename(new_filename)
        shutil.move(full_filename, full_new_filename)
        self.assertEqual(os.path.exists(full_new_filename), True)

    def move_file(self, filename, destination):
        """
        Moves the given file in our fake library to a different destination
        """
        full_filename = self.check_library_filename(filename)
        self.assertEqual(os.path.exists(full_filename), True)

        if destination != '':
            full_destination = self.check_library_filename(destination)
        else:
            full_destination = self.library_path

        # Create the destination dir if it doesn't exist
        os.makedirs(full_destination, exist_ok=True)

        # Now move
        shutil.move(full_filename, full_destination)

        dest_filename = os.path.join(full_destination, os.path.basename(filename))
        self.assertEqual(os.path.exists(dest_filename), True)

    def touch_file(self, filename):
        """
        'touches' a file to have the current modification time, with the extra
        guarantee that the mtime WILL change.  Will set the time one second
        in the future if need be.
        """

        full_filename = self.check_library_filename(filename)

        starting_mtime = int(os.stat(full_filename).st_mtime)

        pathlib.Path(full_filename).touch(exist_ok=True)

        stat_result = os.stat(full_filename)
        ending_mtime = int(stat_result.st_mtime)
        if starting_mtime == ending_mtime:
            new_mtime = ending_mtime + 1
            os.utime(full_filename, times=(stat_result.st_atime, new_mtime))

    def set_file_ancient(self, filename):
        """
        Sets a file's mtime to be as old as it can possibly be.  This is
        only actually used in a single check in AlbumDownloadViewTests, but
        here it is anyway.
        """

        full_filename = self.check_library_filename(filename)
        stat_result = os.stat(full_filename)
        os.utime(full_filename, times=(stat_result.st_atime, 0))

    def age_album(self, artist, album, age_days):
        """
        Given an album already in our database, age it by setting its
        mtime to the current date minus the specified number of days.
        Returns the album object, since it's likely you'll want to use
        it for comparisons anyway.
        """
        album = Album.objects.get(artist__name=artist, name=album)
        album.time_added = timezone.now() - datetime.timedelta(days=age_days)
        album.save()
        return album

    def get_file_contents(self, filename):
        """
        Retrieves file contents from the named file in the library
        """

        full_filename = self.check_library_filename(filename)
        with open(full_filename, 'rb') as df:
            return df.read()

    def set_file_unreadable(self, filename):
        """
        Well set the given file to be unreadable by our add/update processes.
        """

        full_filename = self.check_library_filename(filename)
        os.chmod(full_filename, 0)

    def add_art(self, path='', filename='cover.jpg', basefile='cover_400.jpg'):
        """
        Adds a new cover image to our library in the specified dir.
        """

        self.add_file(basefile, filename, path=path)

    def assertNoErrors(self, appresults):
        """
        Given a list of tuples (as returned from ``App.add()`` or ``App.update()``),
        ensure that none of the lines have status App.STATUS_ERROR
        """
        for (status, line) in appresults:
            self.assertNotEqual(status, App.STATUS_ERROR, msg='Error found: "%s"' % (line))
        return appresults

    def assertErrors(self, appresults, errors_min=1, error=None):
        """
        Given a list of tuples (as returned from ``App.add()`` or ``App.update()``),
        ensure that we have at least ``errors_min`` with a status of App.STATUS_ERROR

        Optionally, also ensure that the errors we DO find match the text specified
        in ``error``.
        """
        error_count = 0
        for (status, line) in appresults:
            if status == App.STATUS_ERROR:
                error_count += 1
                if error is not None:
                    self.assertIn(error, line)
        self.assertGreaterEqual(error_count, errors_min, msg='%d errors expected, %d found' %
            (errors_min, error_count))
        return appresults

    def run_add(self):
        """
        Runs an ``add`` operation on our library, and checks for errors.
        """
        return self.assertNoErrors(list(App.add()))

    def run_add_errors(self, errors_min=1, error=None):
        """
        Runs an ``add`` operation on our library, and expect to see at least
        one error.
        """
        return self.assertErrors(list(App.add()), errors_min, error=error)

    def run_update(self):
        """
        Runs an ``update`` operation on our library, and checks for errors.
        """
        return self.assertNoErrors(list(App.update()))

    def run_update_errors(self, errors_min=1, error=None):
        """
        Runs an ``add`` operation on our library, and expect to see at least
        one error.
        """
        return self.assertErrors(list(App.update()), errors_min, error=error)

# My main Django installation uses an authentication backend of
# django.contrib.auth.backends.RemoteUserBackend, which the test
# client does NOT know what to do with by default.  We could
# have our stuff set a REMOTE_USER variable, perhaps, but it makes
# more sense to just ensure that these tests use a more common
# Django authentication backend.
@override_settings(AUTHENTICATION_BACKENDS=['django.contrib.auth.backends.ModelBackend'])
class ExordiumUserTests(ExordiumTests):
    """
    Tests of ours which also test views, which means that we might need to
    have an admin/staff user in order to test various things.
    """

    @classmethod
    def setUpTestData(self):
        """
        Sets up a user which all tests in this class can use.  The user
        will have staff privileges.
        """
        super(ExordiumUserTests, self).setUpTestData()
        self.user = User.objects.create_user('mainuser')
        self.user.is_staff = True
        self.user.save()

    def login(self):
        """
        Logs in as our staff user
        """
        self.client.force_login(self.user)




from django.test import TestCase

from exordium.models import Artist, Album, Song, App, AlbumArt

class AppModelTests(TestCase):
    """
    Tests against our main App class which don't require our full
    fake library setup.
    """

    def test_ensure_various_artists_create_artist(self):
        """
        Test our App.ensure_various_artists() function and ensure that
        it creates a Various artist if needed.
        """
        self.assertEqual(Artist.objects.count(), 0)
        self.assertEqual(App.ensure_various_artists(), True)
        self.assertEqual(Artist.objects.count(), 1)
        ar = Artist.objects.get()
        self.assertEqual(ar.name, 'Various')
        self.assertEqual(ar.various, True)

    def test_ensure_various_artists_return_existing(self):
        """
        Test our App.ensure_various_artists() function and ensure that
        it returns an existing 'Various' artist if one exists.
        """
        App.ensure_various_artists()
        self.assertEqual(Artist.objects.count(), 1)
        ar = Artist.objects.get()
        ar_pk = ar.pk
        self.assertEqual(ar.name, 'Various')
        self.assertEqual(ar.various, True)

        self.assertEqual(App.ensure_various_artists(), False)
        ar = Artist.objects.get()
        self.assertEqual(ar.pk, ar_pk)
        self.assertEqual(ar.name, 'Various')
        self.assertEqual(ar.various, True)

    def test_add_with_empty_to_add_list(self):
        """
        Test what happens when we run ``App.add()`` with an empty to_add
        list.  Calling this method with a list only happens inside
        ``App.update()``, and that method checks for an empty list before
        calling, so this should never actually happen in "real life,"
        but we'll check it anyway.
        """
        retlines = list(App.add([]))
        self.assertEqual(retlines, [])
        self.assertEqual(Artist.objects.count(), 0)
        self.assertEqual(Album.objects.count(), 0)
        self.assertEqual(Song.objects.count(), 0)




from .base import ExordiumUserTests

from django.urls import reverse

from exordium.models import Artist, Album, Song, App, AlbumArt

class LibraryViewTests(ExordiumUserTests):
    """
    Tests for our main library view index.  Not a whole lot here, honestly.
    I'm not bothering to test the aggregate information shown here, alas.
    """

    def test_without_permission(self):
        """
        Test when we're not actually logged in.
        """

        url = reverse('exordium:library')
        response = self.client.get(url)
        self.assertRedirects(response, '%s?next=%s' % (reverse('admin:login'), url))

    def test_with_permission(self):
        """
        Test when we're logged in.
        """

        self.login()
        response = self.client.get(reverse('exordium:library'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, reverse('exordium:library_update'))

        # We should show our library config, and not have zipfile info
        App.ensure_prefs()
        self.assertContains(response, App.prefs['exordium__base_path'])
        self.assertContains(response, App.prefs['exordium__media_url_html5'])
        self.assertContains(response, App.prefs['exordium__media_url_m3u'])
        self.assertNotContains(response, 'Zipfile Support:</strong> Yes')
        self.assertContains(response, 'Zipfile Support:</strong> No')




from .base import ExordiumTests

from django.urls import reverse

from exordium.models import Artist, Album, Song, App, AlbumArt

class BrowseArtistViewTests(ExordiumTests):
    """
    Tests of our Browse Artists page
    """

    def test_no_artists(self):
        """
        Test the view when there are no artists (except for Various)
        """
        various = Artist.objects.get()

        response = self.client.get(reverse('exordium:browse_artist'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(various)])
        self.assertContains(response, '1 artist')
        self.assertContains(response, reverse('exordium:artist', args=(various.normname,)))

    def test_one_artists(self):
        """
        Test the view when there is one artist (and Various)
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album 1', filename='song1.mp3')
        self.run_add()

        various = Artist.objects.get(normname='various')
        artist = Artist.objects.get(normname='artist')

        response = self.client.get(reverse('exordium:browse_artist'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(artist), repr(various)])
        self.assertContains(response, '2 artists')
        for a in [various, artist]:
            self.assertContains(response, reverse('exordium:artist', args=(a.normname,)))

    def test_ten_artists(self):
        """
        Test the view when there are a bunch of artists
        """
        artists = ['Adipiscing',
            'Amet',
            'Congue',
            'Consectetur',
            'Dolor',
            'Elit',
            'Ipsum',
            'Lorem',
            'Sit',
            'Vivamus',
        ]
        for (idx, artist) in enumerate(artists):
            self.add_mp3(artist=artist, title='Title %d' % (idx+1),
                album='Album', filename='song%d.mp3' % (idx+1))
        self.run_add()

        # add Various to our list, in the properly-sorted place.
        artists.insert(9, 'Various')
        artist_objs = []
        for artist in artists:
            artist_objs.append(Artist.objects.get(name=artist))

        response = self.client.get(reverse('exordium:browse_artist'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data,
            [repr(artist) for artist in artist_objs])
        self.assertContains(response, '11 artists')
        for artist in artist_objs:
            self.assertContains(response, reverse('exordium:artist',
                args=(artist.normname,)))

    def test_classical_song(self):
        """
        Test the view when a single song created multiple artists
        """
        self.add_mp3(artist='Artist', group='Group',
            conductor='Conductor', composer='Composer',
            title='Title', album='Album', filename='song1.mp3')
        self.run_add()

        self.assertEqual(Artist.objects.count(), 5)
        artists = [
            Artist.objects.get(name='Artist'),
            Artist.objects.get(name='Composer'),
            Artist.objects.get(name='Conductor'),
            Artist.objects.get(name='Group'),
            Artist.objects.get(name='Various'),
        ]

        response = self.client.get(reverse('exordium:browse_artist'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data,
            [repr(artist) for artist in artists])
        self.assertContains(response, '5 artists')
        for artist in artists:
            self.assertContains(response, str(artist))
            self.assertContains(response, reverse('exordium:artist',
                args=(artist.normname,)))

    def test_pagination(self):
        """
        Test to make sure that our pagination is working properly.
        The Browse Artists page will show a total of 25 artists
        """
        for num in range(30):
            self.add_mp3(artist='Artist %02d' % (num+1), title='Title %d' % (num+1),
                album='Album %d' % (num+1), filename='song%d.mp3' % (num+1))
        self.run_add()
        self.assertEqual(Artist.objects.count(), 31)

        artists = {}
        for num in range(30):
            artists[num] = Artist.objects.get(name='Artist %02d' % (num+1))

        response = self.client.get(reverse('exordium:browse_artist'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['table'].data), 31)
        for num in range(25):
            self.assertContains(response, '%s<' % (artists[num]))
            self.assertContains(response, reverse('exordium:artist', args=(artists[num].normname,)))
        for num in range(25, 30):
            self.assertNotContains(response, '%s<' % (artists[num]))
            self.assertNotContains(response, reverse('exordium:artist', args=(artists[num].normname,)))
        self.assertNotContains(response, 'Various<')
        self.assertContains(response, '25 of 31 artists')
        self.assertContains(response, '"?page=2"')

        # test page 2
        response = self.client.get(reverse('exordium:browse_artist'), {'page': 2})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['table'].data), 31)
        for num in range(25):
            self.assertNotContains(response, '%s<' % (artists[num]))
            self.assertNotContains(response, reverse('exordium:artist', args=(artists[num].normname,)))
        for num in range(25, 30):
            self.assertContains(response, '%s<' % (artists[num]))
            self.assertContains(response, reverse('exordium:artist', args=(artists[num].normname,)))
        self.assertContains(response, 'Various<')
        self.assertContains(response, '6 of 31 artists')
        self.assertContains(response, '"?page=1"')

    def test_sorting(self):
        """
        Test at least one sort
        """
        self.add_mp3(artist='A Artist', title='Title 1',
            album='Album 1', filename='song1.mp3')
        self.add_mp3(artist='B Artist', title='Title 2',
            album='Album 2', filename='song2.mp3')
        self.run_add()

        various = Artist.objects.get(normname='various')
        artist_a = Artist.objects.get(normname='a artist')
        artist_b = Artist.objects.get(normname='b artist')

        # Initial view, artist name.
        response = self.client.get(reverse('exordium:browse_artist'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data,
            [repr(artist_a), repr(artist_b), repr(various)])
        self.assertContains(response, "?sort=-name")

        # Now sort by name descending
        response = self.client.get(reverse('exordium:browse_artist'), {'sort': '-name'})
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data,
            [repr(various), repr(artist_b), repr(artist_a)])
        self.assertContains(response, "?sort=name")




from .base import ExordiumTests

from django.urls import reverse

from exordium.models import Artist, Album, Song, App, AlbumArt

class AlbumM3UDownloadViewTests(ExordiumTests):
    """
    Tests for our album M3U downloads
    """

    def test_invalid_album(self):
        """
        Tests making a request for an album which can't be found.
        """
        response = self.client.get(reverse('exordium:m3udownload', args=(42,)))
        self.assertEqual(response.status_code, 404)

    def test_one_track_album(self):
        """
        Test getting a one-track album M3U
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album', filename='song1.mp3')
        self.run_add()

        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get()

        self.assertEqual(Song.objects.count(), 1)
        song = Song.objects.get()

        response = self.client.get(reverse('exordium:m3udownload', args=(album.pk,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'audio/mpegurl')
        self.assertEqual(response['Content-Disposition'], 'attachment; filename=Artist_-_Album.m3u')
        self.assertContains(response, '#EXTM3U')
        self.assertContains(response, '#EXTINF:')
        self.assertContains(response, str(song.artist))
        self.assertContains(response, str(song.title))
        self.assertContains(response, '(%s)' % (str(song.album)))
        self.assertContains(response, song.get_download_url_m3u())

    def test_three_track_album(self):
        """
        Test getting a three-track album M3U
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album', filename='song1.mp3')
        self.add_mp3(artist='Artist', title='Title 2',
            album='Album', filename='song2.mp3')
        self.add_mp3(artist='Artist', title='Title 3',
            album='Album', filename='song3.mp3')
        self.run_add()

        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get()

        self.assertEqual(Song.objects.count(), 3)

        response = self.client.get(reverse('exordium:m3udownload', args=(album.pk,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'audio/mpegurl')
        self.assertEqual(response['Content-Disposition'], 'attachment; filename=Artist_-_Album.m3u')
        self.assertContains(response, '#EXTM3U')
        self.assertContains(response, '#EXTINF:')
        self.assertContains(response, '(%s)' % (str(album)))
        for song in Song.objects.all():
            self.assertContains(response, str(song.artist))
            self.assertContains(response, str(song.title))
            self.assertContains(response, song.get_download_url_m3u())




from .base import ExordiumTests

from django.urls import reverse
from django.db.models import Q

from django.templatetags.static import static

from exordium.models import Artist, Album, Song, App, AlbumArt

class BrowseAlbumViewTests(ExordiumTests):
    """
    Tests of our Browse Album page
    """

    def test_no_albums(self):
        """
        Test the view when there are no albums
        """
        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [])

    def test_one_album(self):
        """
        Test the view when there is one album
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album', filename='song1.mp3')
        self.run_add()

        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get()

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(album)])
        self.assertContains(response, '1 album')
        self.assertContains(response, reverse('exordium:album', args=(album.pk,)))
        self.assertContains(response, reverse('exordium:artist', args=(album.artist.normname,)))
        self.assertContains(response, '"%s"' % (static('exordium/no_album_art_small.png')))

    def test_one_album_with_art(self):
        """
        Test the view when there is one album, with album art
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Album', filename='song1.mp3')
        self.add_art()
        self.run_add()

        self.assertEqual(Album.objects.count(), 1)
        album = Album.objects.get()

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(album)])
        self.assertContains(response, '1 album')
        self.assertContains(response, reverse('exordium:album', args=(album.pk,)))
        self.assertContains(response, reverse('exordium:artist', args=(album.artist.normname,)))
        self.assertContains(response, reverse('exordium:albumart', args=(album.pk, 'list',)))

    def test_ordering_three_regular_albums(self):
        """
        Test ordering with three "regular" albums.  Should be alphabetical by
        album name.
        """
        self.add_mp3(artist='Zebra', title='Title 1',
            album='Z Album', filename='song1.mp3')
        self.add_mp3(artist='Artist', title='Title 1',
            album='R Album', filename='song2.mp3')
        self.add_mp3(artist='Zebra', title='Title 1',
            album='Album', filename='song3.mp3')
        self.run_add()

        self.assertEqual(Album.objects.count(), 3)
        albums = [
            Album.objects.get(normname='album'),
            Album.objects.get(normname='r album'),
            Album.objects.get(normname='z album'),
        ]

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(al) for al in albums])
        self.assertContains(response, '3 albums')
        for al in albums:
            self.assertContains(response, reverse('exordium:album', args=(al.pk,)))
            self.assertContains(response, reverse('exordium:artist', args=(al.artist.normname,)))

    def test_ordering_miscellaneous_albums(self):
        """
        By default, "miscellaneous" (non-album) albums should come at
        the end of the list.
        """
        self.add_mp3(artist='Artist', title='Title 1',
            album='Z Album', filename='song1.mp3')
        self.add_mp3(artist='Artist', title='Title 2',
            album='S Album', filename='song2.mp3')
        self.add_mp3(artist='Artist', title='Title 3',
            filename='song3.mp3')
        self.add_mp3(artist='Zebra', title='Title 4',
            album='C Album', filename='song4.mp3')
        self.add_mp3(artist='Zebra', title='Title 5',
            album='V Album', filename='song5.mp3')
        self.add_mp3(artist='Zebra', title='Title 6',
            filename='song6.mp3')
        self.add_mp3(artist='Yellow', title='Title 7',
            filename='song7.mp3')
        self.run_add()

        self.assertEqual(Album.objects.count(), 7)
        albums = []
        for albumname in [
                'C Album',
                'S Album',
                'V Album',
                'Z Album',
                Album.miscellaneous_format_str % ('Artist'),
                Album.miscellaneous_format_str % ('Yellow'),
                Album.miscellaneous_format_str % ('Zebra'),
                ]:
            albums.append(Album.objects.get(name=albumname))

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(al) for al in albums])
        self.assertContains(response, '7 albums')
        for al in albums:
            self.assertContains(response, reverse('exordium:album', args=(al.pk,)))
            self.assertContains(response, reverse('exordium:artist', args=(al.artist.normname,)))

    def test_classical_album(self):
        """
        Test the display of an album with classical tags
        """
        self.add_mp3(artist='Artist', title='Title 1',
            group='Group', conductor='Conductor', composer='Composer',
            album='Album 1', filename='song1.mp3')
        self.run_add()

        self.assertEqual(Artist.objects.count(), 5)
        self.assertEqual(Album.objects.count(), 1)

        album = Album.objects.get()
        artists = []

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(album)])
        self.assertContains(response, 'Album 1')
        self.assertContains(response, reverse('exordium:album', args=(album.pk,)))
        self.assertContains(response, reverse('exordium:artist', args=(album.artist.normname,)))
        for artist in Artist.objects.exclude(name='Various'):
            self.assertContains(response, str(artist))
            self.assertContains(response, reverse('exordium:artist', args=(artist.normname,)))

    def test_classical_album_two_tracks(self):
        """
        Test the display of an album with classical tags.  Two tracks, where
        the artist stays the same but the other classical tags differ.
        """
        self.add_mp3(artist='Artist', title='Title 1',
            group='Group', conductor='Conductor', composer='Composer',
            album='Album 1', filename='song1.mp3')
        self.add_mp3(artist='Artist', title='Title 2',
            group='Group 2', conductor='Conductor 2', composer='Composer 2',
            album='Album 1', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Artist.objects.count(), 8)
        self.assertEqual(Album.objects.count(), 1)

        album = Album.objects.get()
        artists = []

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(album)])
        self.assertContains(response, 'Album 1')
        self.assertContains(response, reverse('exordium:album', args=(album.pk,)))
        self.assertContains(response, reverse('exordium:artist', args=(album.artist.normname,)))
        for artist in Artist.objects.exclude(name='Various'):
            self.assertContains(response, str(artist))
            self.assertContains(response, reverse('exordium:artist', args=(artist.normname,)))

    def test_various_album(self):
        """
        Test display of a Various Artists album
        """
        self.add_mp3(artist='Artist 1', title='Title 1',
            album='Album', filename='song1.mp3')
        self.add_mp3(artist='Artist 2', title='Title 2',
            album='Album', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Artist.objects.count(), 3)
        self.assertEqual(Album.objects.count(), 1)
        artist_1 = Artist.objects.get(name='Artist 1')
        artist_2 = Artist.objects.get(name='Artist 2')
        various = Artist.objects.get(name='Various')
        album = Album.objects.get()

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(album)])
        self.assertContains(response, 'Album')
        self.assertContains(response, reverse('exordium:album', args=(album.pk,)))
        self.assertContains(response, reverse('exordium:artist', args=(album.artist.normname,)))
        self.assertContains(response, reverse('exordium:artist', args=(various.normname,)))
        for artist in [artist_1, artist_2]:
            self.assertNotContains(response, str(artist_1))
            self.assertNotContains(response, reverse('exordium:artist', args=(artist.normname,)))

    def test_classical_album_two_tracks_various(self):
        """
        Test the display of an album with classical tags.  Two tracks, where
        all artist fields change, so it's a Various Artists album.
        """
        self.add_mp3(artist='Artist 1', title='Title 1',
            group='Group', conductor='Conductor', composer='Composer',
            album='Album 1', filename='song1.mp3')
        self.add_mp3(artist='Artist 2', title='Title 2',
            group='Group 2', conductor='Conductor 2', composer='Composer 2',
            album='Album 1', filename='song2.mp3')
        self.run_add()

        self.assertEqual(Artist.objects.count(), 9)
        self.assertEqual(Album.objects.count(), 1)

        album = Album.objects.get()
        artists = []

        various = Artist.objects.get(name='Various')

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(response.context['table'].data, [repr(album)])
        self.assertContains(response, 'Album 1')
        self.assertContains(response, reverse('exordium:album', args=(album.pk,)))
        self.assertContains(response, reverse('exordium:artist', args=(album.artist.normname,)))
        self.assertContains(response, reverse('exordium:artist', args=(various.normname,)))
        for artist in Artist.objects.exclude(Q(name='Artist 1') | Q(name='Artist 2')):
            self.assertContains(response, str(artist))
            self.assertContains(response, reverse('exordium:artist', args=(artist.normname,)))
        for artist in Artist.objects.filter(Q(name='Artist 1') | Q(name='Artist 2')):
            self.assertNotContains(response, str(artist))
            self.assertNotContains(response, reverse('exordium:artist', args=(artist.normname,)))

    def test_pagination(self):
        """
        Test to make sure that our pagination is working properly.
        The album browse page will show a total of 50 albums
        """
        for num in range(60):
            self.add_mp3(artist='Artist', title='Title %d' % (num+1),
                album='Album %02d' % (num+1), filename='song%d.mp3' % (num+1))
        self.run_add()
        self.assertEqual(Album.objects.count(), 60)

        albums = {}
        for num in range(60):
            albums[num] = Album.objects.get(name='Album %02d' % (num+1))

        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['table'].data), 60)
        for num in range(50):
            self.assertContains(response, '%s<' % (albums[num]))
            self.assertContains(response, reverse('exordium:album', args=(albums[num].pk,)))
        for num in range(50, 60):
            self.assertNotContains(response, '%s<' % (albums[num]))
            self.assertNotContains(response, reverse('exordium:album', args=(albums[num].pk,)))
        self.assertContains(response, '50 of 60 albums')
        self.assertContains(response, '"?page=2"')

        # test page 2
        response = self.client.get(reverse('exordium:browse_album'), {'page': 2})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['table'].data), 60)
        for num in range(50):
            self.assertNotContains(response, '%s<' % (albums[num]))
            self.assertNotContains(response, reverse('exordium:album', args=(albums[num].pk,)))
        for num in range(50, 60):
            self.assertContains(response, '%s<' % (albums[num]))
            self.assertContains(response, reverse('exordium:album', args=(albums[num].pk,)))
        self.assertContains(response, '10 of 60 albums')
        self.assertContains(response, '"?page=1"')

    def test_sorting(self):
        """
        Test at least one case of sorting.
        """
        self.add_mp3(artist='Zed', title='Title 1',
            album='Album 1', filename='song1.mp3')
        self.add_mp3(artist='Gee', title='Title 2',
            album='Album 2', filename='song2.mp3')
        self.add_mp3(artist='Artist', title='Title 3',
            album='Album 3', filename='song3.mp3')
        self.run_add()
        self.assertEqual(Album.objects.count(), 3)

        albums = [
            Album.objects.get(name='Album 1'),
            Album.objects.get(name='Album 2'),
            Album.objects.get(name='Album 3'),
        ]
        response = self.client.get(reverse('exordium:browse_album'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['table'].data), 3)
        self.assertQuerysetEqual(response.context['table'].data, [repr(al) for al in albums])
        self.assertContains(response, '"?sort=artist"')

        # test the sorting button
        response = self.client.get(reverse('exordium:browse_album'), {'sort': 'artist'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['table'].data), 3)
        self.assertQuerysetEqual(response.context['table'].data, [repr(al) for al in reversed(albums)])
        self.assertContains(response, '"?sort=-artist"')




from django.test import TestCase

from django.core.management import call_command
from django.core.management.base import CommandError

import io

from exordium.models import Artist, Album, Song, App, AlbumArt

class ImportMysqlAmpacheDatesSubcommandTests(TestCase):
    """
    Tests for our ``importmysqlampachedates`` management subcommand.  Right
    now, just verify that the thing is callable.  Would require having a
    MySQL database available to *actually* test.
    """

    def test_running_command(self):
        out = io.StringIO()
        with self.assertRaises(CommandError) as cm:
            call_command('importmysqlampachedates', stdout=out)

        self.assertIn('the following arguments are required', cm.exception.args[0])
